"""
Pylon.
"""
import os
import sys
sys.path.append(os.path.abspath('..'))
from kerrigan.scv import \
    mkdownloadspath, \
    read_json, write_json, \
    beautify_json_default
from kerrigan.marine import \
    fire_json_default
from kerrigan.tank_msft import \
    initialize_msft_folders_and_secrets, \
    reload_mdm_schema, \
    scan_discrepency_between_capops_and_ztable, \
    deploy_arm_templates, \
    optimize_publish_sproc, \
    upload_fulldc_slotcount, \
    find_out_publish_errors
from kerrigan.ghost import \
    get_secrets

def run(pylon_key):
    menu = _get_menu(pylon_key)
    cmd_to_func = {}
    menu_instructions = []
    for mk, mv in menu.items():
        cmd_to_func[str(mv['id'])] = mk
        menu_instructions.append('{} - {}'.format(mv['id'], mv['description']))
    cmd = ''
    while cmd != 'q':
        cmd = get_cmd(cmd_to_func, menu_instructions)
        func_name = cmd_to_func[cmd]
        func_dict = menu[func_name]
        if 'submenu' not in func_dict:
            globals()[func_name]()
        else:
            param_dict = process_submenu(func_name, func_dict['submenu'])
            globals()[func_name](**param_dict)

def process_submenu(func_name, submenu):
    cmd_set = set()
    for mk, mv in submenu.items():
        cmd_set.add(mk)
    sub_cmd = ''
    while sub_cmd not in cmd_set:
        print('Select running option for {}: '.format(func_name))
        for mk, mv in submenu.items():
            print('\t{} - {}'.format(mk, mv['description']))
        print("Your selection: ", end='')
        sub_cmd = input()
    return submenu[sub_cmd]['param_dict']

def get_cmd(cmd_to_func, menu_instructions):
    for mi in menu_instructions:
        print(mi)
    print("Your choice: ", end='')
    return input()

def quit_menu():
    pass

def _get_menu(secret_key):
    secrets = get_secrets(secret_key)
    return secrets['menu']

if __name__ == "__main__":
    # globals()["process_submenu"]('hello')
    run('msft')