"""
Tank MSFT.
"""
import os
import sys
sys.path.append(os.path.abspath('..'))
import re
import datetime
from kerrigan.scv import \
    mkdir, mkpath, mkdownloadspath, \
    read_json, write_json, \
    read_lines, write_lines, append_line, \
    read_csv_list, read_csv_list_default, \
    write_json_default, \
    query_rows, query_cell, execute_query, \
    get_blob, upload_to_blob, \
    insert_doc, update_doc, \
    find_doc, find_doc_by_id
from kerrigan.marine import \
    run_cmd
from kerrigan.ghost import \
    get_secrets

#region BasicFunctions
def initialize_msft_folders_and_secrets():
    mkdir(mkdownloadspath('resources', 'mdm'))
    mkdir(mkdownloadspath('resources', 'params'))
    mkdir(mkdownloadspath('resources', 'logs'))
    mkdir(mkdownloadspath('resources', 'fulldc'))
    mkdir(mkdownloadspath('resources', 'capernaum'))
    mkdir(mkdownloadspath('resources', 'debug'))
    secrets = read_json(mkpath('.', '..', 'x', 'msft.json'))
    write_json(secrets, mkdownloadspath('resources', 'capernaum', 'msft.json'))
#endregion

#region ZTPFuncions
def find_out_publish_errors():
    """
    Scan errors with given diff.json and error.json.
    """
    error_set = set()
    diffs = read_json(mkdownloadspath(\
        'resources', 'debug', 'diff.json'))['body']['rows']
    error_dict = read_json(mkdownloadspath(\
        'resources', 'debug', 'error.json'))['body']['CodeToPublishErrorMap']
    fields_to_view = get_capops_field_to_view_dict()
    new_diffs = []
    for diff in diffs:
        if diff['RowIdentifier'] in error_dict:
            new_diffs.append(diff)
            for diff_field in diff['DiffFieldModels']:
                if diff_field['FieldName'] in fields_to_view \
                        and diff_field['FieldName'] != 'Plan_LinkedCluster_Code':
                    _query_for_existence(error_set, fields_to_view[diff_field['FieldName']], \
                        diff_field['NewValue'], diff['RowIdentifier'], diff_field['FieldName'])
    error_list = list(error_set)
    write_json(error_list, mkdownloadspath('resources', 'debug', 'publish_report.json'))

def _query_for_existence(error_set, view, target_code, mdm_id, field_name):
    try:
        query_mdm = "SELECT Code from mdm.{} WHERE Code = '{}'".format(view, target_code)
        res = query_cell(_get_mdm_prod_connection_string(), query_mdm)
        if res is None:
            error_msg = '{} - {} - {}'.format(mdm_id, field_name, target_code)
            error_set.add(error_msg)
    except:
        error_msg = '{} - {} - {}'.format(mdm_id, field_name, target_code)
        error_set.add(error_msg)
#endregion

#region ZTable&MDMFuncs
def scan_discrepency_between_capops_and_ztable(\
        start_idx=0, end_idx=6000000, interval=10000):
    log_file_path = \
        mkdownloadspath('resources', 'logs', 'z_mdm_discrepency_{}.log'\
            .format(datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")))
    z_patch_file_path = \
        mkdownloadspath('resources', 'logs', 'z_patch_{}.sql'\
            .format(datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")))
    field_list = get_capops_field_list_no_code_name_newcode()
    for idx in range(start_idx, end_idx, interval):
        if calculate_ztable_capops_checksum(field_list, idx, idx + interval) \
                != calculate_capops_checksum(field_list, idx, idx + interval):
            _calculate_discrepency(field_list, idx, idx + interval, log_file_path, z_patch_file_path)
            print('{} - {}: {}'\
                .format(idx, idx + interval, chr(0x274C)))
        else:
            print('{} - {}: {}'\
                .format(idx, idx + interval, chr(0x2714)))

def _calculate_discrepency(field_list, code_start, code_end, log_file_path, z_patch_file_path):
    # Exact Code-row dict from MDM CapOps dict rows
    mdm_rows_dict = query_capops_fields_dict_rows(field_list, code_start, code_end)
    mdm_dict = {}
    for row in mdm_rows_dict:
        mdm_dict[row['Code']] = row
    # Exact Code-row dict from ZTable CapOps dict rows
    ztable_rows_dict = query_ztable_capops_fields_dict_rows(field_list, code_start, code_end)
    ztable_dict = {}
    for row in ztable_rows_dict:
        ztable_dict[row['Code']] = row
    # Check discrepency for create/update/delete:
    for code, row in mdm_dict.items():
        if code not in ztable_dict:
            # Failed to create in ZTable
            append_line(config_insert_to_ztable_query(code, field_list), z_patch_file_path)
            append_line('Failed to create {} in ZTable.\n'.format(code), log_file_path)
        else:
            # Failed to update in ZTable
            for field_name, field_value in row.items():
                if ztable_dict[code][field_name] != field_value:
                    z_patch_query = "update dbo.ztp_capops set {} = '{}' where Code = '{}'\n"\
                        .format(field_name, field_value, code).replace("'None'", "NULL")
                    append_line(z_patch_query, z_patch_file_path)
                    append_line('Failed to update {} in ZTable.\n'\
                        .format(code), log_file_path)
    for code, row in ztable_dict.items():
        if code not in mdm_dict:
            # Failed to delete in ZTable
            z_patch_query = "delete from dbo.ztp_capops where Code = '{}'\n".format(code)
            append_line(z_patch_query, z_patch_file_path)
            append_line('Failed to delete {} in ZTable.\n'.format(code), log_file_path)
#endregion

#region MDMFunctions
def reload_mdm_schema():
    _initialize_overview_metadata()
    _initialize_capops_dicts()
    _initialize_capops_list()

def _initialize_overview_metadata():
    """
    [
        {
            "EntityName": "CapacityOperationsPlan",
            "StagingBase": "WACAP_CapacityOperationsPlan",
            "ViewName": "WACAP_CapacityOperationsPlan",
            "TableID": "tbl_8_1118_EN",
            "EntityID": 1118,
            "Attributes": [
                {
                    "AttrName": "Priority",
                    "AttrType": 2,
                    "TableColumnID": "uda_1118_4137",
                    "ReferenceTableID": "tbl_8_1127_EN",
                    "ReferenceTableName": "CP_Priority"
                },
                {
                    "AttrName": "State",
                    "AttrType": 2,
                    "TableColumnID": "uda_1118_4138",
                    "ReferenceTableID": "tbl_8_1135_EN",
                    "ReferenceTableName": "CP_State"
                }
            ]
        }
    ]
    """
    overview_query = r"select  e.Name as EntityName, e.StagingBase as StagingBase, s.Name as ViewName, e.EntityTable as TableID, Entity_ID as EntityID from mdm.tblEntity e left join mdm.tblSubscriptionView s on s.Entity_ID = e.ID where e.Model_ID = (SELECT ID FROM mdm.tblModel WHERE Name = 'WACAP')"
    rows = query_rows(_get_mdm_prod_connection_string(), overview_query)
    overview_dict = {}
    for row in rows:
        # Handle exception: WACAP_OrderPriority
        if row['EntityName'] == 'CP_OrderPriority':
            row['ViewName'] = 'WACAP_OrderPriority'
        overview_dict[row['TableID']] = row
    for tableID, row in overview_dict.items():
        query = r"select a.Name as AttrName, a.TableColumn as TableColumnID, a.DomainEntity_ID as ReferenceTableID, a.AttributeType_ID as AttrType from mdm.tblAttribute a join mdm.tblEntity e on a.Entity_ID=e.ID where e.Model_ID=(SELECT ID FROM mdm.tblModel WHERE Name = 'WACAP') and e.Name = '{}'".format(row['EntityName'])
        attribute_rows = query_rows(_get_mdm_prod_connection_string(), query)
        attrs = []
        for attribute_row in attribute_rows:
            if attribute_row['AttrType'] == 2 \
                    and not attribute_row['ReferenceTableID']:
                print('error 1')
            if attribute_row['AttrType'] != 2 \
                    and attribute_row['ReferenceTableID']:
                print('error 2')
            if attribute_row['AttrType'] == 1 \
                    and (attribute_row['AttrName'] == 'Code' \
                        or attribute_row['AttrName'] == 'Name'):
                continue
            if attribute_row['AttrType'] == 3:
                continue
            attr = {}
            # Handle exception: FeasibilityStatus
            attr['AttrName'] = attribute_row['AttrName'] \
                    if attribute_row['AttrType'] != 2 \
                        or attribute_row['AttrName'] == 'FeasibilityStatus' \
                    else '{}_Code'.format(attribute_row['AttrName'])
            attr['AttrType'] = attribute_row['AttrType']
            attr['TableColumnID'] = attribute_row['TableColumnID']
            if attribute_row['AttrType'] == 2:
                attr['ReferenceTableID'] = _get_ref_table(\
                    attribute_row['ReferenceTableID'])[0]
                attr['ReferenceTableName'] = _get_ref_table(\
                    attribute_row['ReferenceTableID'])[1]
            # Add EntityID, EntityName and ViewName to attr dict.
            if attribute_row['AttrType'] == 2:
                attr['EntityID'] = attribute_row['ReferenceTableID']
                attr['EntityName'] = overview_dict[attr['ReferenceTableID']]['EntityName']
                attr['ViewName'] = overview_dict[attr['ReferenceTableID']]['ViewName']
            attrs.append(attr)
        row['Attributes'] = attrs
    write_json(overview_dict, mkdownloadspath('resources', 'mdm', 'mdm_overview.json'))
    return overview_dict

def _initialize_capops_dicts():
    """
    EntityID_to_Field.json
    EntityID_to_EntityName.json
    EntityID_to_ViewName.json
    Field_to_ViewName.json
    """
    overview_dict = read_json(mkdownloadspath('resources', 'mdm', 'mdm_overview.json'))
    entityid_to_field = {}
    entityid_to_entityname = {}
    entityid_to_viewname = {}
    field_to_viewname = {}
    for attr in overview_dict['tbl_8_1118_EN']['Attributes']:
        if attr['AttrType'] == 2:
            # if attr['EntityID'] in entityid_to_field:
            #     print(attr['EntityID'])
            entityid_to_field[attr['EntityID']] = attr['AttrName']
            entityid_to_entityname[attr['EntityID']] = attr['EntityName']
            entityid_to_viewname[attr['EntityID']] = attr['ViewName']
            field_to_viewname[attr['AttrName']] = attr['ViewName']
    write_json(entityid_to_field, mkdownloadspath('resources', 'mdm', 'EntityID_to_Field.json'))
    write_json(entityid_to_entityname, mkdownloadspath('resources', 'mdm', 'EntityID_to_EntityName.json'))
    write_json(entityid_to_viewname, mkdownloadspath('resources', 'mdm', 'EntityID_to_ViewName.json'))
    write_json(field_to_viewname, mkdownloadspath('resources', 'mdm', 'Field_to_ViewName.json'))

def _initialize_capops_list():
    """
    CapOps_Field_List.json
    """
    overview_dict = read_json(mkdownloadspath('resources', 'mdm', 'mdm_overview.json'))
    capops_fields = []
    for attr in overview_dict['tbl_8_1118_EN']['Attributes']:
        capops_fields.append(attr['AttrName'])
    write_json(capops_fields, mkdownloadspath('resources', 'mdm', 'CapOps_Field_List.json'))

def calculate_capops_checksum(field_list, code_start, code_end):
    """
    Return the checksum value of CapOps table,
    with given fields and code's range.
    """
    query_fields_body = ','.join(field_list)
    query_body = 'SELECT CHECKSUM_AGG(CHECKSUM({})) FROM mdm.wacap_CapacityOperationsPlan WITH(NOLOCK) WHERE Code >= {} AND Code < {}'\
        .format(query_fields_body, code_start, code_end)
    return query_cell(_get_mdm_prod_connection_string(), query_body)

def query_capops_fields_dict_rows(field_list, code_start, code_end):
    """
    Return list of dict for CapOps fields' info,
    with given field list and code's range.
    """
    query_fields_body = ','.join(field_list)
    query_body = 'SELECT Code, {} FROM mdm.wacap_CapacityOperationsPlan WITH(NOLOCK) WHERE Code >= {} AND Code < {}'\
        .format(query_fields_body, code_start, code_end)
    return query_rows(_get_mdm_prod_connection_string(), query_body, code_end - code_start + 1)

def config_insert_to_ztable_query(code, field_list):
    """
    Return an insert to ZTable query based on the code and fields.
    """
    query_fields_body = ','.join(field_list)
    query_body = "SELECT Code, {} FROM mdm.wacap_CapacityOperationsPlan WITH(NOLOCK) WHERE Code = '{}'"\
        .format(query_fields_body, code)
    query_result = query_rows(_get_mdm_prod_connection_string(), query_body, 1)
    mdm_row = query_result[0]
    field_names = []
    field_values = []
    for field_name, field_value in mdm_row.items():
        if field_value:
            field_names.append(field_name)
            field_values.append("'{}'".format(field_value))
    z_patch_query = "INSERT INTO [dbo].[ZTP_CapOps] ({}) Values({})\n"\
        .format(','.join(field_names), ','.join(field_values))
    return z_patch_query

def get_capops_field_list_no_code_name_newcode():
    """
    Return CapOps fields, yet no 'Code', 'Name' or 'NewCode'
    """
    return read_json(mkdownloadspath('resources', 'mdm', 'CapOps_Field_List.json'))

def get_capops_field_to_view_dict():
    """
    Return CapOps field to view name dict.
    """
    return read_json(mkdownloadspath('resources', 'mdm', 'Field_to_ViewName.json'))

def _get_ref_table(ref_table):
    query = r"select  e.Name as EntityName, s.Name as ViewName, e.EntityTable as TableID, * from mdm.tblEntity e left join mdm.tblSubscriptionView s on s.Entity_ID = e.ID where e.Model_ID = (SELECT ID FROM mdm.tblModel WHERE Name = 'WACAP') and e.ID = {}"\
        .format(ref_table)
    rows = query_rows(_get_mdm_prod_connection_string(), query)
    if len(rows) != 1:
        print(ref_table)
    return (rows[0]['TableID'], rows[0]['EntityName'])

def optimize_publish_sproc(column_name, column_id, \
        source_sproc_path, target_sproc_path):
    sproc_optimized = read_lines(source_sproc_path)
    for i in range(len(sproc_optimized) - 1, -1, -1):
        if sproc_optimized[i].startswith('--z-optimization'):
            optimization_type = sproc_optimized[i].split('~')[1].strip()
            if optimization_type == 'nullline':
                sproc_optimized[i - 1] = _config_nullline(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'longline':
                sproc_optimized[i - 1] = _config_longline(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'shortline':
                sproc_optimized[i - 1] = _config_shortline(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'newuda':
                sproc_optimized[i - 1] = _config_newuda(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'normaluda':
                sproc_optimized[i - 1] = _config_normaluda(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'nullif':
                sproc_optimized[i - 1] = _config_nullif(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'casewhenlong':
                sproc_optimized[i - 1] = _config_casewhenlong(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
            elif optimization_type == 'casewhenshort':
                sproc_optimized[i - 1] = _config_casewhenshort(sproc_optimized[i - 1].replace('\n', ''), column_name, column_id)
    write_lines(sproc_optimized, target_sproc_path)

def _config_nullline(line, column_name, column_id):
    middle = 'CONVERT(NVARCHAR(MAX), inserted.'
    p1 = line.split(middle)[0]
    p1_update = p1 + 'NULL , '
    line = line.replace(p1.strip(), p1_update.strip())
    line = line + ', CONVERT(NVARCHAR(MAX), inserted.[uda_1118_{}]) '.format(column_id) + '\n'
    return line

def _config_longline(line, column_name, column_id):
    middle = 'CONVERT(NVARCHAR(MAX), inserted.'
    p1 = line.split(middle)[0]
    p1_update = p1 + 'CONVERT(NVARCHAR(MAX), deleted.[uda_1118_{}]) , '.format(column_id)
    line = line.replace(p1.strip(), p1_update.strip())
    line = line + ', CONVERT(NVARCHAR(MAX), inserted.[uda_1118_{}]) '.format(column_id) + '\n'
    return line

def _config_newuda(line, column_name, column_id):
    return line + '\n' + '            ,New_uda_1118_{} NVARCHAR(MAX) NULL '.format(column_id) + '\n'

def _config_normaluda(line, column_name, column_id):
    return line + '\n' + '            ,uda_1118_{} NVARCHAR(MAX) NULL '.format(column_id) + '\n'

def _config_shortline(line, column_name, column_id):
    return line + '\n' + ',[uda_1118_{}] --[{}] '.format(column_id, column_name) + '\n'

def _config_nullif(line, column_name, column_id):
    return line + ',[uda_1118_{}] = NULLIF(stgl.[{}],@NULLText)'.format(column_id, column_name) + '\n'

def _config_casewhenlong(line, column_name, column_id):
    return '                                          END ,[uda_1118_{}] = CASE WHEN stgl.[{}] = @NULLText THEN NULL\n                                               WHEN stgl.[{}] IS NULL THEN en.[uda_1118_{}]\n                                               ELSE stgl.[{}]\n                                          END\n'.format(column_id, column_name, column_name, column_id, column_name)

def _config_casewhenshort(line, column_name, column_id):
    return line + '\n' + 'NULLIF(stgl.[{}],@NULLText),'.format(column_name) + '\n'
#endregion

#region ZTableFunctions
def calculate_ztable_capops_checksum(field_list, code_start, code_end):
    """
    Return the checksum value of ZTable's CapOps table,
    with given fields and code's range.
    """
    query_fields_body = ','.join(field_list)
    query_body = 'SELECT CHECKSUM_AGG(CHECKSUM({})) FROM ZTP_CapOps WITH(NOLOCK) WHERE Code >= {} AND Code < {}'\
        .format(query_fields_body, code_start, code_end)
    return query_cell(_get_ztable_prod_connection_string(), query_body)

def query_ztable_capops_fields_dict_rows(field_list, code_start, code_end):
    """
    Return list of dict for ZTable's CapOps fields' info,
    with given field list and code's range.
    """
    query_fields_body = ','.join(field_list)
    query_body = 'SELECT Code, {} FROM ZTP_CapOps WITH(NOLOCK) WHERE Code >= {} AND Code < {}'\
        .format(query_fields_body, code_start, code_end)
    return query_rows(_get_ztable_prod_connection_string(), query_body, code_end - code_start + 1)
#endregion

#region BlobFunctions
def get_json_from_ztp_blob(container, blob):
    msft_keys = get_secrets('msft')
    account_name = msft_keys['secrets']['users']['ztp_blob_prod']
    account_key = msft_keys['secrets']['passwords']['ztp_blob_prod']
    return get_blob(account_name, account_key, container, blob)

def upload_json_to_ztp_blob(container, blob, file_path):
    msft_keys = get_secrets('msft')
    account_name = msft_keys['secrets']['users']['ztp_blob_prod']
    account_key = msft_keys['secrets']['passwords']['ztp_blob_prod']
    return upload_to_blob(account_name, account_key, \
        container, blob, file_path)
#endregion

#region CosmosFunctions
def query_doc(db_name, collection_name, query, \
        options=None, partition_key=None):
    """
    Example:
        query = {
            'query': \
                'SELECT * FROM x WHERE x.data.auditData.RowIdentifier="2000002"',
            'parameters': [{'name':'@id', \
        }
        res = query_doc('ztpcycles', 'auditLogs', query, \
            {'partitionKey': 'mdmpublish', 'enableCrossPartitionQuery': True})
    """ 
    cosmos_secret = _get_ztp_prod_cosmos_host_and_key()
    return find_doc(cosmos_secret['host'], cosmos_secret['key'], \
        db_name, collection_name, query, \
        options, partition_key)

def query_doc_by_id(db_name, collection_name, doc_id, options=None):
    """
    Example:
        res = query_doc_by_id('ztpcycles', 'auditLogs', \
            'f11a78aa-1fd0-4652-8154-eaf918a4bfb3', \
            {'partitionKey': 'mdmpublish', 'enableCrossPartitionQuery': True})
    """
    cosmos_secret = _get_ztp_prod_cosmos_host_and_key()
    return find_doc_by_id(cosmos_secret['host'], cosmos_secret['key'], \
        db_name, collection_name, doc_id, options)
#endregion

#region AzureCLIFunctions
def deploy_arm_templates(env):
    azure_dict = _get_azure_dict()
    templates = azure_dict['templates']
    sub = azure_dict[env]['sub']
    rg = azure_dict[env]['rg']
    for _, template_path in templates.items():
        param_path = _config_temp_param_file(template_path, \
            azure_dict['param_template'], \
            azure_dict[env]['param_dict'])
        cmd = azure_dict['cli']['deploy_arm_template']\
            .replace('_sub', sub)\
            .replace('_rg', rg)\
            .replace('_template', template_path)\
            .replace('_param', param_path)
        run_cmd(cmd.split(' '))

def _config_temp_param_file(template_path, param_template, params):
    template = read_json(template_path)
    require_param_dict = {}
    full_param_dict = _collect_full_param_dict(params)
    for pk,_ in template['parameters'].items():
        require_param_dict[pk] = full_param_dict[pk]
    param_template['parameters'] = require_param_dict
    param_path = mkdownloadspath('resources', 'params', 'param_temp.json')
    write_json(param_template, param_path)
    return param_path

def _collect_full_param_dict(param_set_dict):
    param_dict = {}
    for _, v in param_set_dict.items():
        for pk, pv in v.items():
            param_dict[pk] = pv
    return param_dict
#endregion

#region FullDCFitout
def upload_fulldc_slotcount():
    """Read from local dc.csv, process and upload to blob."""
    dc_list = read_csv_list(mkdownloadspath('resources', 'fulldc', 'dc.csv'))
    dc_dict_list = _convert_dc_csv_lines_to_dict_list(dc_list)
    full_dc_table = {
        'ref_table': _generate_ref_table_from_dc_dict_list(dc_dict_list),
        'slot_table': _generate_slot_table_from_dc_dict_list(dc_dict_list)
    }
    file_path = mkdownloadspath('resources', 'fulldc', 'dc.json')
    write_json(full_dc_table, file_path)
    upload_json_to_ztp_blob('z-functions-output', 'dc.json', file_path)

def _convert_dc_csv_lines_to_dict_list(lines):
    """Convert csv lines to dict list."""
    dc_field_map = \
    {
        'segmention': 'number',
        'footprint': 'number',
        'reservedkw': 'number',
        'adjustedkw': 'number',
        'leftkw': 'number',
        'segmenation-footprint-reserved': 'number',
        '1*10': 'number',
        '1*20': 'number',
        'currentdate': 'datetime',
        'FFCR': 'datetime',
    }
    row_list = []
    clean_header_string = re.sub('[^A-Za-z0-9,*-]+', '', lines[0])
    headers = clean_header_string.split(',')
    for i in range(1, len(lines)):
        row_dict = {}
        row_values = lines[i].split(',')
        for j in range(0, len(row_values)):
            if not row_values[j] or row_values[j].lower() == 'null':
                continue
            if headers[j] in dc_field_map and \
                    dc_field_map[headers[j]] == 'number':
                row_values[j] = float(row_values[j]) \
                    if '.' in row_values[j] else int(row_values[j])
            if headers[j] in dc_field_map and \
                    dc_field_map[headers[j]] == 'datetime':
                pass
            row_dict[headers[j]] = row_values[j]
        # Ignore rows with no info at csv, aka formatting/manual input mistake.
        if len(row_dict) > 1:
            row_list.append(row_dict)
    return row_list

def _generate_ref_table_from_dc_dict_list(dc_dict_list):
    """Generate ref table for full dc input."""
    ref_table = []
    region_ref_table = {}
    region_dc_ref_table = {}
    region_az_ref_table = {}
    region_dc_az_ref_table = {}
    for dc_dict in dc_dict_list:
        region = dc_dict['MetroValue']
        dc = dc_dict['DC']
        az = _extract_az_from_dc_dict(dc_dict)
        colo = dc_dict['colocation']
        region_dc = '{}:{}'.format(region, dc)
        region_az = '{}:{}'.format(region, az)
        region_dc_az = '{}:{}:{}'.format(region, dc, az)
        dc_colo = '{}:{}'.format(dc, colo)
        if region not in region_ref_table:
            region_ref_table[region] = set()
        region_ref_table[region].add(dc_colo)
        if region_dc not in region_dc_ref_table:
            region_dc_ref_table[region_dc] = set()
        region_dc_ref_table[region_dc].add(dc_colo)
        if region_az not in region_az_ref_table:
            region_az_ref_table[region_az] = set()
        region_az_ref_table[region_az].add(dc_colo)
        if region_dc_az not in region_dc_az_ref_table:
            region_dc_az_ref_table[region_dc_az] = set()
        region_dc_az_ref_table[region_dc_az].add(dc_colo)
    _append_to_ref_table(ref_table, region_ref_table)
    _append_to_ref_table(ref_table, region_dc_ref_table)
    _append_to_ref_table(ref_table, region_az_ref_table)
    _append_to_ref_table(ref_table, region_dc_az_ref_table)
    return ref_table

def _generate_slot_table_from_dc_dict_list(dc_dict_list):
    """Generate slot count table for full dc input."""
    slot_table = []
    for dc_dict in dc_dict_list:
        slot_dict = {
            'Colo': dc_dict['colocation'],
            'DcShortCode': dc_dict['DC'],
            'ColoStatus': dc_dict['type'],
            'Eg': dc_dict['EG'],
            'RemainingPower': dc_dict['adjustedkw'],
            'AvailableDate': dc_dict['FFCR'],
            'C10': _extract_slot_count_from_dc_dict(\
                '1*10', dc_dict),
            'C20': _extract_slot_count_from_dc_dict(\
                '1*20', dc_dict),
            'LastUpdateDate': dc_dict['currentdate']
        }
        slot_table.append(slot_dict)
    return slot_table

def _extract_slot_count_from_dc_dict(slot_key, dc_dict):
    if slot_key not in dc_dict:
        return 999 if dc_dict['type'] == 'Power Feed' else 0
    return dc_dict[slot_key]

def _extract_az_from_dc_dict(dc_dict):
    if 'AZ' not in dc_dict or dc_dict['AZ'].lower() == 'non-az':
        return 'unknown'
    return dc_dict['AZ']

def _append_to_ref_table(ref_table, ref_dict):
    for rowkey, dc_colos in ref_dict.items():
        ref = {
            'RowKey': rowkey,
            'DcColos': ','.join(list(dc_colos))
        }
        ref_table.append(ref)
#endregion

#region PrivateHelpers
def _get_mdm_prod_connection_string():
    msft_keys = get_secrets('msft')
    return msft_keys['secrets']['connection_strings']['mdm_prod']

def _get_ztable_prod_connection_string():
    msft_keys = get_secrets('msft')
    return msft_keys['secrets']['connection_strings']['ztable_prod']

def _get_azure_dict():
    msft_keys = get_secrets('msft')
    return msft_keys['azure']

def _get_ztp_prod_cosmos_host_and_key():
    msft_keys = get_secrets('msft')
    return \
    {
        'host': msft_keys['secrets']['users']['ztp_cosmos_prod'], 
        'key': msft_keys['secrets']['passwords']['ztp_cosmos_prod']
    }
#endregion

if __name__ == "__main__":
    run_tank_msft()